@@ -1,5 +1,5 @@
 {-# LANGUAGE BangPatterns, CPP, DeriveDataTypeable, MagicHash #-}
-{-# OPTIONS_GHC -funbox-strict-fields #-}
+{-# OPTIONS_GHC -fno-full-laziness -funbox-strict-fields #-}
 
 module Data.HashMap.Base
     (
@@ -36,6 +36,7 @@ module Data.HashMap.Base
       -- * Difference and intersection
     , difference
     , intersection
+    , intersectionWith
 
       -- * Folds
     , foldl'
@@ -69,14 +70,14 @@ module Data.HashMap.Base
     , two
     , unionArrayBy
     , update16
-    , update16'
-    , update16With
+    , update16M
+    , update16With'
     , updateOrConcatWith
     ) where
 
 import Control.Applicative ((<$>), Applicative(pure))
 import Control.DeepSeq (NFData(rnf))
-import Control.Monad.ST (ST, runST)
+import Control.Monad.ST (ST)
 import Data.Bits ((.&.), (.|.), complement)
 import qualified Data.Foldable as Foldable
 import qualified Data.List as L
@@ -89,6 +90,7 @@ import qualified Data.HashMap.Array as A
 import qualified Data.Hashable as H
 import Data.Hashable (Hashable)
 import Data.HashMap.PopCount (popCount)
+import Data.HashMap.Unsafe (runST)
 import Data.HashMap.UnsafeShift (unsafeShiftL, unsafeShiftR)
 import Data.Typeable (Typeable)
 
@@ -103,6 +105,7 @@ hash :: H.Hashable a => a -> Hash
 hash = fromIntegral . H.hash
 
 data Leaf k v = L !k v
+  deriving (Eq)
 
 instance (NFData k, NFData v) => NFData (Leaf k v) where
     rnf (L k v) = rnf k `seq` rnf v
@@ -149,9 +152,21 @@ instance (Show k, Show v) => Show (HashMap k v) where
 instance Traversable (HashMap k) where
     traverse f = traverseWithKey (const f)
 
--- NOTE: This is just a placeholder.
 instance (Eq k, Eq v) => Eq (HashMap k v) where
-    a == b = toList a == toList b
+    (==) = equal
+
+equal :: (Eq k, Eq v) => HashMap k v -> HashMap k v -> Bool
+equal (BitmapIndexed b1 ary1) (BitmapIndexed b2 ary2) =
+    b1 == b2 && A.toList ary1 == A.toList ary2
+equal (Leaf k1 v1) (Leaf k2 v2) =
+    k1 == k2 && v1 == v2
+equal (Full ary1) (Full ary2) =
+    A.toList ary1 == A.toList ary2
+equal (Collision k1 ary1) (Collision k2 ary2) =
+    k1 == k2 && A.length ary1 == A.length ary2
+      && L.null (A.toList ary1 L.\\ A.toList ary2)
+equal Empty Empty = True
+equal _     _     = False
 
 ------------------------------------------------------------------------
 -- * Construction
@@ -254,42 +269,38 @@ bitmapIndexedOrFull b ary
 -- key in this map.  If this map previously contained a mapping for
 -- the key, the old value is replaced.
 insert :: (Eq k, Hashable k) => k -> v -> HashMap k v -> HashMap k v
-insert k0 v0 m0 = runST (go h0 k0 v0 0 m0)
+insert k0 v0 m0 = go h0 k0 v0 0 m0
   where
     h0 = hash k0
-    go !h !k x !_ Empty = return $! Leaf h (L k x)
+    go !h !k x !_ Empty = Leaf h (L k x)
     go h k x s t@(Leaf hy l@(L ky y))
         | hy == h = if ky == k
                     then if x `ptrEq` y
-                         then return t
-                         else return $! Leaf h (L k x)
-                    else return $! collision h l (L k x)
-        | otherwise = two s h k x hy ky y
+                         then t
+                         else Leaf h (L k x)
+                    else collision h l (L k x)
+        | otherwise = runST (two s h k x hy ky y)
     go h k x s t@(BitmapIndexed b ary)
-        | b .&. m == 0 = do
-            !ary' <- A.insert' ary i $! Leaf h (L k x)
-            return $! bitmapIndexedOrFull (b .|. m) ary'
-        | otherwise = do
-            st <- A.index_ ary i
-            !st' <- go h k x (s+bitsPerSubkey) st
-            if st' `ptrEq` st
-                then return t
-                else do
-                    ary' <- A.update' ary i st'
-                    return $! BitmapIndexed b ary'
+        | b .&. m == 0 =
+            let ary' = A.insert ary i $! Leaf h (L k x)
+            in bitmapIndexedOrFull (b .|. m) ary'
+        | otherwise =
+            let !st  = A.index ary i
+                !st' = go h k x (s+bitsPerSubkey) st
+            in if st' `ptrEq` st
+               then t
+               else BitmapIndexed b (A.update ary i st')
       where m = mask h s
             i = sparseIndex b m
-    go h k x s t@(Full ary) = do
-        st <- A.index_ ary i
-        !st' <- go h k x (s+bitsPerSubkey) st
-        if st' `ptrEq` st
-            then return t
-            else do
-                ary' <- update16' ary i st'
-                return $! Full ary'
+    go h k x s t@(Full ary) =
+        let !st  = A.index ary i
+            !st' = go h k x (s+bitsPerSubkey) st
+        in if st' `ptrEq` st
+            then t
+            else Full (update16 ary i st')
       where i = index h s
     go h k x s t@(Collision hy v)
-        | h == hy   = return $! Collision h (updateOrSnocWith const k x v)
+        | h == hy   = Collision h (updateOrSnocWith const k x v)
         | otherwise = go h k x s $ BitmapIndexed (mask hy s) (A.singleton t)
 #if __GLASGOW_HASKELL__ >= 700
 {-# INLINABLE insert #-}
@@ -310,19 +321,19 @@ unsafeInsert k0 v0 m0 = runST (go h0 k0 v0 0 m0)
         | otherwise = two s h k x hy ky y
     go h k x s t@(BitmapIndexed b ary)
         | b .&. m == 0 = do
-            ary' <- A.insert' ary i $! Leaf h (L k x)
+            ary' <- A.insertM ary i $! Leaf h (L k x)
             return $! bitmapIndexedOrFull (b .|. m) ary'
         | otherwise = do
-            st <- A.index_ ary i
+            st <- A.indexM ary i
             st' <- go h k x (s+bitsPerSubkey) st
-            A.unsafeUpdate' ary i st'
+            A.unsafeUpdateM ary i st'
             return t
       where m = mask h s
             i = sparseIndex b m
     go h k x s t@(Full ary) = do
-        st <- A.index_ ary i
+        st <- A.indexM ary i
         st' <- go h k x (s+bitsPerSubkey) st
-        A.unsafeUpdate' ary i st'
+        A.unsafeUpdateM ary i st'
         return t
       where i = index h s
     go h k x s t@(Collision hy v)
@@ -338,8 +349,8 @@ two = go
   where
     go s h1 k1 v1 h2 k2 v2
         | bp1 == bp2 = do
-            st <- go (s+bitsPerSubkey) h1 k1 v1 h2 k2 v2 
-            ary <- A.singleton' st
+            st <- go (s+bitsPerSubkey) h1 k1 v1 h2 k2 v2
+            ary <- A.singletonM st
             return $! BitmapIndexed bp1 ary
         | otherwise  = do
             mary <- A.new 2 $ Leaf h1 (L k1 v1)
@@ -362,34 +373,34 @@ two = go
 -- >   where f new old = new + old
 insertWith :: (Eq k, Hashable k) => (v -> v -> v) -> k -> v -> HashMap k v
             -> HashMap k v
-insertWith f k0 v0 m0 = runST (go h0 k0 v0 0 m0)
+insertWith f k0 v0 m0 = go h0 k0 v0 0 m0
   where
     h0 = hash k0
-    go !h !k x !_ Empty = return $! Leaf h (L k x)
+    go !h !k x !_ Empty = Leaf h (L k x)
     go h k x s (Leaf hy l@(L ky y))
         | hy == h = if ky == k
-                    then return $! Leaf h (L k (f x y))
-                    else return $! collision h l (L k x)
-        | otherwise = two s h k x hy ky y
+                    then Leaf h (L k (f x y))
+                    else collision h l (L k x)
+        | otherwise = runST (two s h k x hy ky y)
     go h k x s (BitmapIndexed b ary)
-        | b .&. m == 0 = do
-            ary' <- A.insert' ary i $! Leaf h (L k x)
-            return $! bitmapIndexedOrFull (b .|. m) ary'
-        | otherwise = do
-            st <- A.index_ ary i
-            st' <- go h k x (s+bitsPerSubkey) st
-            ary' <- A.update' ary i st'
-            return $! BitmapIndexed b ary'
+        | b .&. m == 0 =
+            let ary' = A.insert ary i $! Leaf h (L k x)
+            in bitmapIndexedOrFull (b .|. m) ary'
+        | otherwise =
+            let st   = A.index ary i
+                st'  = go h k x (s+bitsPerSubkey) st
+                ary' = A.update ary i $! st'
+            in BitmapIndexed b ary'
       where m = mask h s
             i = sparseIndex b m
-    go h k x s (Full ary) = do
-        st <- A.index_ ary i
-        st' <- go h k x (s+bitsPerSubkey) st
-        ary' <- update16' ary i st'
-        return $! Full ary'
+    go h k x s (Full ary) =
+        let st   = A.index ary i
+            st'  = go h k x (s+bitsPerSubkey) st
+            ary' = update16 ary i $! st'
+        in Full ary'
       where i = index h s
     go h k x s t@(Collision hy v)
-        | h == hy   = return $! Collision h (updateOrSnocWith f k x v)
+        | h == hy   = Collision h (updateOrSnocWith f k x v)
         | otherwise = go h k x s $ BitmapIndexed (mask hy s) (A.singleton t)
 #if __GLASGOW_HASKELL__ >= 700
 {-# INLINABLE insertWith #-}
@@ -409,19 +420,19 @@ unsafeInsertWith f k0 v0 m0 = runST (go h0 k0 v0 0 m0)
         | otherwise = two s h k x hy ky y
     go h k x s t@(BitmapIndexed b ary)
         | b .&. m == 0 = do
-            ary' <- A.insert' ary i $! Leaf h (L k x)
+            ary' <- A.insertM ary i $! Leaf h (L k x)
             return $! bitmapIndexedOrFull (b .|. m) ary'
         | otherwise = do
-            st <- A.index_ ary i
+            st <- A.indexM ary i
             st' <- go h k x (s+bitsPerSubkey) st
-            A.unsafeUpdate' ary i st'
+            A.unsafeUpdateM ary i st'
             return t
       where m = mask h s
             i = sparseIndex b m
     go h k x s t@(Full ary) = do
-        st <- A.index_ ary i
+        st <- A.indexM ary i
         st' <- go h k x (s+bitsPerSubkey) st
-        A.unsafeUpdate' ary i st'
+        A.unsafeUpdateM ary i st'
         return t
       where i = index h s
     go h k x s t@(Collision hy v)
@@ -434,54 +445,48 @@ unsafeInsertWith f k0 v0 m0 = runST (go h0 k0 v0 0 m0)
 -- | /O(log n)/ Remove the mapping for the specified key from this map
 -- if present.
 delete :: (Eq k, Hashable k) => k -> HashMap k v -> HashMap k v
-delete k0 m0 = runST (go h0 k0 0 m0)
+delete k0 m0 = go h0 k0 0 m0
   where
     h0 = hash k0
-    go !_ !_ !_ Empty = return Empty
+    go !_ !_ !_ Empty = Empty
     go h k _ t@(Leaf hy (L ky _))
-        | hy == h && ky == k = return Empty
-        | otherwise          = return t
+        | hy == h && ky == k = Empty
+        | otherwise          = t
     go h k s t@(BitmapIndexed b ary)
-        | b .&. m == 0 = return t
-        | otherwise = do
+        | b .&. m == 0 = t
+        | otherwise =
             let !st = A.index ary i
-            !st' <- go h k (s+bitsPerSubkey) st
-            if st' `ptrEq` st
-                then return t
+                !st' = go h k (s+bitsPerSubkey) st
+            in if st' `ptrEq` st
+                then t
                 else case st' of
-                Empty | A.length ary == 1 -> return Empty
-                      | otherwise -> do
-                          ary' <- A.delete' ary i
-                          return $! BitmapIndexed (b .&. complement m) ary'
-                _ -> do
-                    ary' <- A.update' ary i st'
-                    return $! BitmapIndexed b ary'
+                Empty | A.length ary == 1 -> Empty
+                      | otherwise -> BitmapIndexed (b .&. complement m) (A.delete ary i)
+                _ -> BitmapIndexed b (A.update ary i st')
       where m = mask h s
             i = sparseIndex b m
-    go h k s t@(Full ary) = do
-        let !st = A.index ary i
-        !st' <- go h k (s+bitsPerSubkey) st
-        if st' `ptrEq` st
-            then return t
+    go h k s t@(Full ary) =
+        let !st   = A.index ary i
+            !st' = go h k (s+bitsPerSubkey) st
+        in if st' `ptrEq` st
+            then t
             else case st' of
-            Empty -> do
-                ary' <- A.delete' ary i
-                let bm = fullNodeMask .&. complement (1 `unsafeShiftL` i)
-                return $! BitmapIndexed bm ary'
-            _ -> do
-                ary' <- A.update' ary i st'
-                return $! Full ary'
+            Empty ->
+                let ary' = A.delete ary i
+                    bm   = fullNodeMask .&. complement (1 `unsafeShiftL` i)
+                in BitmapIndexed bm ary'
+            _ -> Full (A.update ary i st')
       where i = index h s
     go h k _ t@(Collision hy v)
         | h == hy = case indexOf k v of
             Just i
                 | A.length v == 2 ->
                     if i == 0
-                    then return $! Leaf h (A.index v 1)
-                    else return $! Leaf h (A.index v 0)
-                | otherwise -> return $! Collision h (A.delete v i)
-            Nothing -> return t
-        | otherwise = return t
+                    then Leaf h (A.index v 1)
+                    else Leaf h (A.index v 0)
+                | otherwise -> Collision h (A.delete v i)
+            Nothing -> t
+        | otherwise = t
 #if __GLASGOW_HASKELL__ >= 700
 {-# INLINABLE delete #-}
 #endif
@@ -573,7 +578,7 @@ unionWith f = go 0
         | b1 .&. m2 == 0 = let ary' = A.insert ary1 i t2
                                b'   = b1 .|. m2
                            in bitmapIndexedOrFull b' ary'
-        | otherwise      = let ary' = A.updateWith ary1 i $ \st1 ->
+        | otherwise      = let ary' = A.updateWith' ary1 i $ \st1 ->
                                    go (s+bitsPerSubkey) st1 t2
                            in BitmapIndexed b1 ary'
         where
@@ -584,7 +589,7 @@ unionWith f = go 0
         | b2 .&. m1 == 0 = let ary' = A.insert ary2 i $! t1
                                b'   = b2 .|. m1
                            in bitmapIndexedOrFull b' ary'
-        | otherwise      = let ary' = A.updateWith ary2 i $ \st2 ->
+        | otherwise      = let ary' = A.updateWith' ary2 i $ \st2 ->
                                    go (s+bitsPerSubkey) t1 st2
                            in BitmapIndexed b2 ary'
       where
@@ -594,12 +599,12 @@ unionWith f = go 0
     go s (Full ary1) t2 =
         let h2   = leafHashCode t2
             i    = index h2 s
-            ary' = update16With ary1 i $ \st1 -> go (s+bitsPerSubkey) st1 t2
+            ary' = update16With' ary1 i $ \st1 -> go (s+bitsPerSubkey) st1 t2
         in Full ary'
     go s t1 (Full ary2) =
         let h1   = leafHashCode t1
             i    = index h1 s
-            ary' = update16With ary2 i $ \st2 -> go (s+bitsPerSubkey) t1 st2
+            ary' = update16With' ary2 i $ \st2 -> go (s+bitsPerSubkey) t1 st2
         in Full ary'
 
     leafHashCode (Leaf h _) = h
@@ -623,20 +628,20 @@ unionArrayBy f b1 b2 ary1 ary2 = A.run $ do
     mary <- A.new_ (popCount b')
     -- iterate over nonzero bits of b1 .|. b2
     -- it would be nice if we could shift m by more than 1 each time
-    let hasBit b m = b .&. m /= 0
+    let ba = b1 .&. b2
         go !i !i1 !i2 !m
-            | m > b'                     = return ()
-            | hasBit b1 m && hasBit b2 m = do
+            | m > b'        = return ()
+            | b' .&. m == 0 = go i i1 i2 (m `unsafeShiftL` 1)
+            | ba .&. m /= 0 = do
                 A.write mary i $! f (A.index ary1 i1) (A.index ary2 i2)
                 go (i+1) (i1+1) (i2+1) (m `unsafeShiftL` 1)
-            | hasBit b1 m                = do
-                A.write mary i =<< A.index_ ary1 i1
+            | b1 .&. m /= 0 = do
+                A.write mary i =<< A.indexM ary1 i1
                 go (i+1) (i1+1) (i2  ) (m `unsafeShiftL` 1)
-            | hasBit b2 m                = do
-                A.write mary i =<< A.index_ ary2 i2
+            | otherwise     = do
+                A.write mary i =<< A.indexM ary2 i2
                 go (i+1) (i1  ) (i2+1) (m `unsafeShiftL` 1)
-            | otherwise                  = go i i1 i2 (m `unsafeShiftL` 1)
-    go 0 0 0 1
+    go 0 0 0 (b' .&. negate b') -- XXX: b' must be non-zero
     return mary
     -- TODO: For the case where b1 .&. b2 == b1, i.e. when one is a
     -- subset of the other, we could use a slightly simpler algorithm,
@@ -665,6 +670,9 @@ map f = go
                            A.map' (\ (L k v) -> L k (f v)) ary
 {-# INLINE map #-}
 
+-- TODO: We should be able to use mutation to create the new
+-- 'HashMap'.
+
 -- | /O(n)/ Transform this map by accumulating an Applicative result
 -- from every value.
 traverseWithKey :: Applicative f => (k -> v1 -> f v2) -> HashMap k v1
@@ -682,7 +690,7 @@ traverseWithKey f = go
 ------------------------------------------------------------------------
 -- * Difference and intersection
 
--- | /O(n+m)/ Difference of two maps. Return elements of the first map
+-- | /O(n*log m)/ Difference of two maps. Return elements of the first map
 -- not existing in the second.
 difference :: (Eq k, Hashable k) => HashMap k v -> HashMap k w -> HashMap k v
 difference a b = foldlWithKey' go empty a
@@ -694,7 +702,7 @@ difference a b = foldlWithKey' go empty a
 {-# INLINABLE difference #-}
 #endif
 
--- | /O(n+m)/ Intersection of two maps. Return elements of the first
+-- | /O(n*log m)/ Intersection of two maps. Return elements of the first
 -- map for keys existing in the second.
 intersection :: (Eq k, Hashable k) => HashMap k v -> HashMap k w -> HashMap k v
 intersection a b = foldlWithKey' go empty a
@@ -706,6 +714,20 @@ intersection a b = foldlWithKey' go empty a
 {-# INLINABLE intersection #-}
 #endif
 
+-- | /O(n+m)/ Intersection of two maps. If a key occurs in both maps
+-- the provided function is used to combine the values from the two
+-- maps.
+intersectionWith :: (Eq k, Hashable k) => (v1 -> v2 -> v3) -> HashMap k v1
+                 -> HashMap k v2 -> HashMap k v3
+intersectionWith f a b = foldlWithKey' go empty a
+  where
+    go m k v = case lookup k b of
+                 Just w -> insert k (f v w) m
+                 _      -> m
+#if __GLASGOW_HASKELL__ >= 700
+{-# INLINABLE intersectionWith #-}
+#endif
+
 ------------------------------------------------------------------------
 -- * Folds
 
@@ -951,12 +973,12 @@ updateOrConcatWith f ary1 ary2 = A.run $ do
           | i2 >= n2 = return ()
           | otherwise = case A.index indices i2 of
                Just i1 -> do -- key occurs in both arrays, store combination in position i1
-                             L k v1 <- A.index_ ary1 i1
-                             L _ v2 <- A.index_ ary2 i2
+                             L k v1 <- A.indexM ary1 i1
+                             L _ v2 <- A.indexM ary2 i2
                              A.write mary i1 (L k (f v1 v2))
                              go iEnd (i2+1)
                Nothing -> do -- key is only in ary2, append to end
-                             A.write mary iEnd =<< A.index_ ary2 i2
+                             A.write mary iEnd =<< A.indexM ary2 i2
                              go (iEnd+1) (i2+1)
     go n1 0
     return mary
@@ -969,21 +991,21 @@ updateOrConcatWith f ary1 ary2 = A.run $ do
 
 -- | /O(n)/ Update the element at the given position in this array.
 update16 :: A.Array e -> Int -> e -> A.Array e
-update16 ary idx b = runST (update16' ary idx b)
+update16 ary idx b = runST (update16M ary idx b)
 {-# INLINE update16 #-}
 
 -- | /O(n)/ Update the element at the given position in this array.
-update16' :: A.Array e -> Int -> e -> ST s (A.Array e)
-update16' ary idx b = do
+update16M :: A.Array e -> Int -> e -> ST s (A.Array e)
+update16M ary idx b = do
     mary <- clone16 ary
     A.write mary idx b
     A.unsafeFreeze mary
-{-# INLINE update16' #-}
+{-# INLINE update16M #-}
 
 -- | /O(n)/ Update the element at the given position in this array, by applying a function to it.
-update16With :: A.Array e -> Int -> (e -> e) -> A.Array e
-update16With ary idx f = update16 ary idx $! f (A.index ary idx)
-{-# INLINE update16With #-}
+update16With' :: A.Array e -> Int -> (e -> e) -> A.Array e
+update16With' ary idx f = update16 ary idx $! f (A.index ary idx)
+{-# INLINE update16With' #-}
 
 -- | Unsafely clone an array of 16 elements.  The length of the input
 -- array is not checked.
@@ -993,22 +1015,22 @@ clone16 ary =
     A.thaw ary 0 16
 #else
     do mary <- A.new_ 16
-       A.index_ ary 0 >>= A.write mary 0
-       A.index_ ary 1 >>= A.write mary 1
-       A.index_ ary 2 >>= A.write mary 2
-       A.index_ ary 3 >>= A.write mary 3
-       A.index_ ary 4 >>= A.write mary 4
-       A.index_ ary 5 >>= A.write mary 5
-       A.index_ ary 6 >>= A.write mary 6
-       A.index_ ary 7 >>= A.write mary 7
-       A.index_ ary 8 >>= A.write mary 8
-       A.index_ ary 9 >>= A.write mary 9
-       A.index_ ary 10 >>= A.write mary 10
-       A.index_ ary 11 >>= A.write mary 11
-       A.index_ ary 12 >>= A.write mary 12
-       A.index_ ary 13 >>= A.write mary 13
-       A.index_ ary 14 >>= A.write mary 14
-       A.index_ ary 15 >>= A.write mary 15
+       A.indexM ary 0 >>= A.write mary 0
+       A.indexM ary 1 >>= A.write mary 1
+       A.indexM ary 2 >>= A.write mary 2
+       A.indexM ary 3 >>= A.write mary 3
+       A.indexM ary 4 >>= A.write mary 4
+       A.indexM ary 5 >>= A.write mary 5
+       A.indexM ary 6 >>= A.write mary 6
+       A.indexM ary 7 >>= A.write mary 7
+       A.indexM ary 8 >>= A.write mary 8
+       A.indexM ary 9 >>= A.write mary 9
+       A.indexM ary 10 >>= A.write mary 10
+       A.indexM ary 11 >>= A.write mary 11
+       A.indexM ary 12 >>= A.write mary 12
+       A.indexM ary 13 >>= A.write mary 13
+       A.indexM ary 14 >>= A.write mary 14
+       A.indexM ary 15 >>= A.write mary 15
        return mary
 #endif
 
