@@ -1,4 +1,4 @@
-{-# LANGUAGE GADTs #-}
+{-# LANGUAGE CPP, GADTs #-}
 
 module Main where
 
@@ -134,7 +134,7 @@ main = do
 
           -- Folds
         , bench "foldl'" $ whnf (HM.foldl' (+) 0) hmi
-        , bench "foldr" $ whnf (HM.foldr (:) []) hmi
+        , bench "foldr" $ nf (HM.foldr (:) []) hmi
 
           -- Filter
         , bench "filter" $ whnf (HM.filter (\ v -> v .&. 1 == 0)) hmi
@@ -149,9 +149,39 @@ main = do
 
           -- fromList
         , bgroup "fromList"
-          [ bench "String" $ whnf HM.fromList elems
-          , bench "ByteString" $ whnf HM.fromList elemsBS
-          , bench "Int" $ whnf HM.fromList elemsI
+          [ bgroup name
+            [ bgroup "long"
+              [ bench "String" $ whnf fl1 elems
+              , bench "ByteString" $ whnf fl2 elemsBS
+              , bench "Int" $ whnf fl3 elemsI
+              ]
+            , bgroup "short"
+              [ bench "String" $ whnf fl1 elemsDup
+              , bench "ByteString" $ whnf fl2 elemsDupBS
+              , bench "Int" $ whnf fl3 elemsDupI
+              ]
+            ]
+          | (name,fl1,fl2,fl3)
+               <- [("Base",HM.fromList,HM.fromList,HM.fromList)
+                  ,("insert",fromList_insert,fromList_insert,fromList_insert)]
+          ]
+          -- fromList
+        , bgroup "fromListWith"
+          [ bgroup name
+            [ bgroup "long"
+              [ bench "String" $ whnf (fl1 (+)) elems
+              , bench "ByteString" $ whnf (fl2 (+)) elemsBS
+              , bench "Int" $ whnf (fl3 (+)) elemsI
+              ]
+            , bgroup "short"
+              [ bench "String" $ whnf (fl1 (+)) elemsDup
+              , bench "ByteString" $ whnf (fl2 (+)) elemsDupBS
+              , bench "Int" $ whnf (fl3 (+)) elemsDupI
+              ]
+            ]
+          | (name,fl1,fl2,fl3)
+               <- [("Base",HM.fromListWith,HM.fromListWith,HM.fromListWith)
+                  ,("insert",fromListWith_insert,fromListWith_insert,fromListWith_insert)]
           ]
         ]
   where
@@ -170,6 +200,13 @@ main = do
     keysBS'  = UBS.rnd' 8 n
     keysI'   = UI.rnd' (n+n) n
 
+    keysDup    = US.rnd 2 n
+    keysDupBS  = UBS.rnd 2 n
+    keysDupI   = UI.rnd (n`div`4) n
+    elemsDup   = zip keysDup [1..n]
+    elemsDupBS = zip keysDupBS [1..n]
+    elemsDupI  = zip keysDupI [1..n]
+
 ------------------------------------------------------------------------
 -- * HashMap
 
@@ -230,3 +267,18 @@ insertIM xs m0 = foldl' (\m (k, v) -> IM.insert k v m) m0 xs
 
 deleteIM :: [Int] -> IM.IntMap Int -> IM.IntMap Int
 deleteIM xs m0 = foldl' (\m k -> IM.delete k m) m0 xs
+
+------------------------------------------------------------------------
+-- * Reference implementations
+
+fromList_insert :: (Eq k, Hashable k) => [(k, v)] -> HM.HashMap k v
+fromList_insert = foldl' (\ m (k, v) -> HM.insert k v m) HM.empty
+#if __GLASGOW_HASKELL__ >= 700
+{-# INLINABLE fromList_insert #-}
+#endif
+
+fromListWith_insert :: (Eq k, Hashable k) => (v -> v -> v) -> [(k, v)] -> HM.HashMap k v
+fromListWith_insert f = foldl' (\ m (k, v) -> HM.insertWith f k v m) HM.empty
+#if __GLASGOW_HASKELL__ >= 700
+{-# INLINABLE fromListWith_insert #-}
+#endif
